"""Event dispatching and listening.

This module provides a simple `EventTarget` which can be used
to dispatch events.
The easiest way to use it is to inherit from it.

Attributes:
    EventHandler ((Event) -> `Any`): (Type alias) Callable which takes an event. If the callable is a coroutine it is awaited.
    EventFilter ((Event) -> `bool`): (Type alias) Callable which takes an event and returns a `bool`.
        This is used to "filter" events.
    EventSpecifierType (Union[str, Type[NamedEvent]]): (Type alias) Type of an object which conveys the name of an `Event`.
        Usually this is just a string with the name of the event, however you can
        also use a `NamedEvent` class.
"""
import abc
import asyncio
import inspect
import logging
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Coroutine, Dict, Iterable, List, Mapping, MutableMapping, Optional, Set, \
    Tuple, Type, TypeVar, Union, \
    overload

import lettercase

__all__ = ["Event", "NamedEvent",
           "EventErrorEvent",
           "EventHandler", "EventFilter",
           "OneTimeEventListener", "EventListener",
           "EventSpecifierType", "resolve_event_specifier",
           "EventTarget"]

log = logging.getLogger(__name__)


class Event:
    """Base type of all events.

    Args:
        name: Name of the event
        arg: Acts like kwargs, but doesn't need to be unpacked, thus being more
            efficient. This argument is optional. kwargs values will override
            the values in this mapping.
        **kwargs: Attributes to set on the event.

    Attributes:
        name (str): Name of the event.

    Dynamic attributes (those passed to the constructor) can also be accessed
    using the `get` method.
    """
    name: str

    _keys: Set[str]

    @overload
    def __init__(self, name: str, arg: Mapping[str, Any], **kwargs: Any) -> None:
        ...

    @overload
    def __init__(self, name: str, arg: Iterable[Tuple[str, Any]], **kwargs: Any) -> None:
        ...

    @overload
    def __init__(self, name: str, **kwargs: Any) -> None:
        ...

    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        self.name = name

        body = dict(*args, **kwargs)
        self.__dict__.update(body)

        try:
            keys = self._keys
        except AttributeError:
            keys = self._keys = set()

        keys.update(body.keys())

    def __repr__(self) -> str:
        type_name = type(self).__name__
        name = self.name

        attrs_str = ", ".join(self._keys)
        return f"{type_name} ({name}) {{{attrs_str}}}"

    def get(self, key: str, *, default: Any = None) -> Optional[Any]:
        """Get the value of a key.

        Unlike `getattr`, the key must have been passed to the event.
        If you need to get any attribute, use `getattr` instead.

        Args:
            key: Name of the key to retrieve
            default: Default value to return if the key doesn't exist.
                (Default is `None`)
        """
        if key in self._keys:
            return getattr(self, key, default=default)
        return default


def _trim_suffix(s: str, suffix: str) -> str:
    """Return the string without the given suffix.

    Args:
        s: String to remove suffix from
        suffix: Exact suffix to remove
    """
    if s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s


class NamedEvent(Event, abc.ABC):
    """Event which automatically gets its name from its class.

    Args:
        **kwargs: Attributes to set on the event.
            (Passed to the `Event` constructor)

    This is useful for subclasses of `Event` whose name is used for the event name.
    You may also manually set the event name by setting `__event_name__` on the class.

    This is an abstract class, you cannot use it directly. You can either use an `Event`
    or subclass this class.
    """
    __event_name__: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(self.get_event_name(), *args, **kwargs)

    @classmethod
    def _create_name(cls) -> str:
        """Generate the name for the event representing this class.

        This uses the class name and converts it to snake_case. ("NamedEvent" -> "named_event")
        If the class name endswith "Event" it is stripped. So "NamedEvent" would become "named".

        Returns:
            Generated name
        """
        cls_name = _trim_suffix(cls.__name__, "Event")

        name = lettercase.convert_to(cls_name, lettercase.LetterCase.SNAKE)
        cls.__event_name__ = name
        return name

    @classmethod
    def get_event_name(cls) -> str:
        """Get the event name for the event representing this class."""
        try:
            return cls.__event_name__
        except AttributeError:
            return cls._create_name()


EventHandler = Callable[[Event], Any]


class EventErrorEvent(Event):
    """Event type passed to `EventTarget.on_event_error`.

    Attributes:
        handler (EventHandler): Callback that received the event.
        exception (Exception): Error that was raised by the handler.
        original_event (Event): Event that was to be dispatched but caused an error.
    """
    handler: EventHandler
    exception: Exception
    original_event: Event

    def __init__(self, handler: EventHandler, exception: Exception, original_event: Event) -> None:
        super().__init__("event_error", handler=handler, exception=exception, original_event=original_event)


EventFilter = Callable[[Event], bool]


@dataclass()
class OneTimeEventListener:
    """One-time event listener.

    Special event listener which waits for the first event matching the conditions and then removes itself.

    Attributes:
        future (Future): Future that was returned to the callee and is used
            to communicate.
            When an event occurs which meets the conditions it is set
            as the result of this future. If an error occurs, it is
            set as the exception.
        condition (Optional[EventFilter]): Callback which takes an `Event` and returns `True`
            to accept the event and `False` otherwise.
            If not provided, all events are accepted.
    """
    future: asyncio.Future
    condition: Optional[EventFilter] = None


@dataclass()
class EventListener:
    """Persistent event listener.

    Attributes:
        handler (EventHandler): Callback which will be called whenever an event occurs.
    """
    handler: EventHandler


T = TypeVar("T")


def _push_map_list(mapping: MutableMapping[str, List[T]], key: str, item: T) -> None:
    try:
        items = mapping[key]
    except KeyError:
        mapping[key] = [item]
    else:
        items.append(item)


EventSpecifierType = Union[str, Type[NamedEvent]]


def resolve_event_specifier(event: EventSpecifierType) -> str:
    """Resolve an `EventSpecifierType` to the name of the event.

    Raises:
        TypeError: If something other than an `EventSpecifierType` is passed.
    """
    if isinstance(event, str):
        return event
    elif issubclass(event, NamedEvent):
        return event.get_event_name()
    else:
        raise TypeError(f"Unknown event specifier type {type(event).__name__!r}: {event}")


def safe_wait_for(fut: Awaitable, timeout: float = None, *, loop: asyncio.AbstractEventLoop = None) -> asyncio.Future:
    """Same as `asyncio.wait_for` but returns `None` instead of raising an `asyncio.TimeoutError`."""

    async def wrapper() -> Optional[Any]:
        try:
            return await asyncio.wait_for(fut, timeout, loop=loop)
        except asyncio.TimeoutError:
            return None

    return asyncio.ensure_future(wrapper(), loop=loop)


class EventTarget:
    """An object that can dispatch `Event` instances to listeners.

    Args:
        loop: You can specify the loop which will be used for various asynchronous actions.
            If you don't specify a loop, it will be dynamically retrieved when needed.
    """
    _one_time_listeners: Dict[str, List[OneTimeEventListener]]
    _listeners: Dict[str, List[EventListener]]
    _children: Set["EventTarget"]

    _loop: Optional[asyncio.AbstractEventLoop]

    def __init__(self, *, loop: asyncio.AbstractEventLoop = None) -> None:
        self._loop = loop
        self._one_time_listeners = {}
        self._listeners = {}
        self._children = set()

    def __str__(self) -> str:
        total_listeners = len(self._one_time_listeners) + len(self._listeners)
        total_children = len(self._children)
        return f"{type(self).__name__}(listeners: {total_listeners}, children: {total_children})"

    @overload
    def wait_for(self, event: EventSpecifierType, *,
                 check: EventFilter = None) -> Awaitable[Event]:
        ...

    @overload
    def wait_for(self, event: EventSpecifierType, *,
                 check: EventFilter = None,
                 timeout: float) -> Awaitable[Optional[Event]]:
        ...

    def wait_for(self, event: EventSpecifierType, *,
                 check: EventFilter = None,
                 timeout: float = None) -> Awaitable[Optional[Event]]:
        """Wait for an event to be dispatched.

        This is similar to adding a listener, but it only listens once and
        returns the event that was dispatched instead of calling a callback.

        Args:
            event: Name of the event to listen to.
                Also accepts `NamedEvent` types.
            check: Callback which takes an `Event` and returns `True`
                to accept the event and `False` otherwise.
                If not provided, all events are accepted.
            timeout: Amount of seconds to wait before aborting.
                If not provided it will wait forever.

        Returns:
            `Future` which when awaited yields the event.
            If a timeout was given the result will be `None`
            if it timed-out.
        """
        event = resolve_event_specifier(event)

        # not using self._loop.create_future because loop may be None
        future = asyncio.Future(loop=self._loop)
        listener = OneTimeEventListener(future, check)

        _push_map_list(self._one_time_listeners, event, listener)

        return safe_wait_for(future, timeout=timeout, loop=self._loop)

    def add_dispatcher(self, target: "EventTarget") -> None:
        """Add a child event target which will dispatch the same events.

        Args:
            target: Event target to add as a child.
        """
        self._children.add(target)

    def remove_dispatcher(self, target: "EventTarget") -> None:
        """Remove a child event target.

        Args:
            target: Event target to remove.
        """

        self._children.discard(target)

    def add_listener(self, event: EventSpecifierType, listener: Union[EventHandler, EventListener]) -> EventListener:
        """Add a listener for an `Event` which is called whenever said event is dispatched.

        Args:
            event: Name of the event to listen to.
                Also accepts `NamedEvent` types.
            listener: Listener to add.

        Returns:
            The `EventListener` that was added.
            If the provided listener already is an `EventListener`,
            the return value is the same value.
        """
        event = resolve_event_specifier(event)

        if not isinstance(listener, EventListener):
            listener = EventListener(listener)

        _push_map_list(self._listeners, event, listener)
        return listener

    def remove_listener(self, event: EventSpecifierType, target: Union[EventHandler, EventListener]) -> bool:
        """Remove a previously added event listener.

        When a `EventHandler` is passed, all event listeners with the given handler are removed.

        Args:
            event: Name of the event for which to remove a listener.
                Also accepts `NamedEvent` types.
            target: Listener to remove.

        Returns:
            bool: Whether the listener was successfully removed.
        """
        event = resolve_event_specifier(event)

        try:
            listeners = self._listeners[event]
        except KeyError:
            return False

        if isinstance(target, EventListener):
            try:
                listeners.remove(target)
                return True
            except ValueError:
                return False

        removed_some: bool = False
        for i, listener in reversed(listeners):
            if listener.handler == target:
                del listeners[i]
                removed_some = True

        return removed_some

    @overload
    def has_listener(self, event: EventSpecifierType) -> bool:
        ...

    @overload
    def has_listener(self, event: EventSpecifierType,
                     target: Union[EventHandler, EventListener, OneTimeEventListener]) -> bool:
        ...

    def has_listener(self, event: EventSpecifierType,
                     target: Union[EventHandler, EventListener, OneTimeEventListener] = None) -> bool:
        """Check whether an event has a listener.

        This includes one-time listeners, but does not check method
        listeners (on_<event> methods defined in the event target class).

        Args:
            event: Name of the event to check.
                Also accepts `NamedEvent` types.
            target: Listener to check whether it has been added.
                If this is `None` the method checks whether the
                event has **any** listeners.
        """
        event = resolve_event_specifier(event)

        if target is None:
            return bool(self._listeners.get(event)) or bool(self._one_time_listeners.get(event))
        elif isinstance(target, OneTimeEventListener):
            try:
                listeners = self._one_time_listeners[event]
            except KeyError:
                return False
            else:
                return target in listeners
        else:
            try:
                listeners = self._listeners[event]
            except KeyError:
                return False

            if isinstance(target, EventListener):
                return target in listeners
            else:
                for listener in listeners:
                    if listener.handler == target:
                        return True

                return False

    def dispatch(self, event: Event) -> Optional[asyncio.Future]:
        """Dispatch an event.

        Events are first propagated to the "one-time listeners" (i.e. those added by `wait_for`),
        then the instances "on_<event>" methods are called and finally
        the listeners.

        Events are also propagated to child event targets.

        If an error occurs while dispatching an event, `on_event_error` is called.

        Args:
            event: Event to be dispatched.

        Returns:
            Optional[Future]: Future aggregating all asynchronous dispatches.
                Can be awaited to make sure all handlers have been executed.
        """
        if log.isEnabledFor(logging.DEBUG):
            # converting event to string may be expensive, so only
            # do it if we want it logged
            log.debug(f"dispatching event in {self}: {event}")

        futures: List[Union[Coroutine, asyncio.Future]] = []
        event_name = event.name

        one_time_listeners = self._one_time_listeners.get(event_name)
        if one_time_listeners:
            to_remove: List[int] = []

            for i, listener in enumerate(one_time_listeners):
                future = listener.future
                if future.cancelled():
                    to_remove.append(i)
                    continue

                condition = listener.condition
                if condition:
                    try:
                        res = condition(event)
                    except Exception as e:
                        res = False
                        future.set_exception(e)
                        to_remove.append(i)
                else:
                    res = True

                if res:
                    future.set_result(event)
                    to_remove.append(i)

            if len(to_remove) == len(one_time_listeners):
                one_time_listeners.clear()
            else:
                for i in reversed(to_remove):
                    del one_time_listeners[i]

        try:
            method = getattr(self, f"on_{event_name}")
        except AttributeError:
            pass
        else:
            futures.append(self._run_event(method, event))

        listeners = self._listeners.get(event_name)
        if listeners:
            for listener in listeners:
                futures.append(self._run_event(listener.handler, event))

        for child_target in self._children:
            fut = child_target.dispatch(event)
            if fut is not None:
                futures.append(asyncio.ensure_future(fut, loop=self._loop))

        if futures:
            return asyncio.gather(*futures, loop=self._loop)
        else:
            return None

    async def _run_event(self, method: EventHandler, event: Event):
        try:
            if inspect.iscoroutine(method):
                await method(event)
            else:
                method(event)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            error_event = EventErrorEvent(method, e, event)

            with suppress(asyncio.CancelledError):
                await self.on_event_error(error_event)

    # noinspection PyMethodMayBeStatic
    async def on_event_error(self, event: EventErrorEvent) -> None:
        """Called when an error occurs during event dispatching.

        Args:
            event: Special event type which contains some meta information
                next to the original event.
        """
        log.exception(f"There was an error when {event.handler} handled {event.original_event}: {event.exception}")



"""Player models.

Attributes:
    MixerMap (Dict[str, MixerPlayer]): (Type alias) Mapping from player id to `MixerPlayer`
"""

import abc
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional, cast

from andesite.transform import RawDataType, from_centi, from_milli, map_build_all_values_from_raw, \
    map_build_values_from_raw, map_convert_values, \
    map_convert_values_from_milli, map_convert_values_to_milli, to_centi, to_milli, transform_input
from .filters import FilterMap

__all__ = ["PlayerFrameStats", "BasePlayer", "MixerPlayer", "MixerMap", "Player"]


@dataclass
class PlayerFrameStats:
    """Frame statistics of a player."""
    loss: int
    success: int
    usable: bool


@dataclass
class BasePlayer(abc.ABC):
    """Abstract class for Andesite players.

    See Also:

        - `Player`
        - `MixerPlayer`

    Attributes:
        time (datetime): current utc datetime on the node
        position (Optional[float]): position of the current playing track, or `None` if nothing is playing
        paused (bool): whether or not the player is paused
        volume (float): the volume of the player
        filters (FilterMap): map of filter name -> filter settings for each filter present
        frame (PlayerFrameStats) Player frame stats
    """
    time: datetime
    position: Optional[float]
    paused: bool
    volume: float
    filters: FilterMap
    frame: PlayerFrameStats

    @property
    def live_position(self) -> Optional[float]:
        """Interpolated version of :py:attr:`position` based on the time that has passed.

        Returns:
            `None` if there is no position attribute,
            equal to position if the player is paused, and interpolated
            otherwise.
        """
        if not self.position:
            return None

        if self.paused:
            return self.position

        return self.position + (datetime.utcnow() - self.time).total_seconds()

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values_from_milli(data, "position")
        map_convert_values(data, volume=from_centi)
        map_build_values_from_raw(data, filters=FilterMap, frame=PlayerFrameStats)

        time_float = from_milli(int(data["time"]))
        data["time"] = datetime.utcfromtimestamp(time_float)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values_to_milli(data, "position")
        map_convert_values(data, volume=to_centi)
        data["time"] = to_milli(cast(datetime, data["time"]).timestamp())


@dataclass
class MixerPlayer(BasePlayer):
    """Player used by `Player` as a mixer player.

    See Also:
        `BasePlayer`
    """
    ...


MixerMap = Dict[str, MixerPlayer]


@dataclass
class Player(BasePlayer):
    """

    Attributes:
        mixer (MixerMap): map of mixer player id -> mixer player
        mixer_enabled (bool): whether or not the mixer is the current source of audio
    """
    mixer: MixerMap
    mixer_enabled: bool

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        data = transform_input(super(), data)
        map_build_all_values_from_raw(data["mixer"], MixerPlayer)
        return data



"""Event models for the web socket client."""

import logging
from typing import Any, Dict, Generic, TYPE_CHECKING, TypeVar

from .event_target import NamedEvent
from .models import AndesiteEvent, ReceiveOperation

if TYPE_CHECKING:
    from .web_socket_client import AbstractAndesiteWebSocketClient

__all__ = ["WebSocketConnectEvent", "WebSocketDisconnectEvent",
           "RawMsgReceiveEvent", "MsgReceiveEvent",
           "RawMsgSendEvent"]

ROPT = TypeVar("ROPT", bound=ReceiveOperation)
ET = TypeVar("ET", bound=AndesiteEvent)

log = logging.getLogger(__name__)


class WebSocketConnectEvent(NamedEvent):
    """Event dispatched when a connection has been established.

    Attributes:
        client (AbstractAndesiteWebSocketClient): Web socket client which
            connected.
    """
    __event_name__ = "ws_connect"

    client: "AbstractAndesiteWebSocketClient"

    def __init__(self, client: "AbstractAndesiteWebSocketClient") -> None:
        super().__init__(client=client)


class WebSocketDisconnectEvent(NamedEvent):
    """Event dispatched when a client was disconnected.

    Attributes:
        client (AbstractAndesiteWebSocketClient): Web socket client which connected.
        deliberate (bool): Whether the disconnect was deliberate.
    """
    __event_name__ = "ws_disconnect"

    client: "AbstractAndesiteWebSocketClient"
    deliberate: bool

    def __init__(self, client: "AbstractAndesiteWebSocketClient", deliberate: bool) -> None:
        super().__init__(client=client, deliberate=deliberate)


class RawMsgReceiveEvent(NamedEvent):
    """Event emitted when a web socket message is received.

    Attributes:
        client (AbstractAndesiteWebSocket): Web socket client that
            received the message.
        body (Dict[str, Any]): Raw body of the received message.
            Note: The body isn't manipulated in any way other
            than being loaded from the raw JSON string.
            For example, the names are still in dromedaryCase.
    """
    client: "AbstractAndesiteWebSocketClient"
    body: Dict[str, Any]

    def __init__(self, client: "AbstractAndesiteWebSocketClient", body: Dict[str, Any]) -> None:
        super().__init__(client=client, body=body)


class MsgReceiveEvent(NamedEvent, Generic[ROPT]):
    """Event emitted when a web socket message is received.

    Attributes:
        client (AbstractAndesiteWebSocketClient): Web socket client that
            received the message.
        op (str): Operation.
            This will be one of the following:

            - connection-id
            - player-update
            - stats
            - event
        data (ReceiveOperation): Loaded message model.
            The type of this depends on the op.
    """
    client: "AbstractAndesiteWebSocketClient"
    op: str
    data: ROPT

    def __init__(self, client: "AbstractAndesiteWebSocketClient", op: str, data: ROPT) -> None:
        super().__init__(client=client, op=op, data=data)


class RawMsgSendEvent(NamedEvent):
    """Event dispatched before a web socket message is sent.

    It's important to note that this is not a receipt of a message being sent,
    this event is dispatched even if the message fails to send.

    Attributes:
        client (AbstractAndesiteWebSocketClient): Web socket client that
            received the message.
        guild_id (int): guild id
        op (str): Op-code to be executed
        body (Dict[str, Any]): Raw body of the message
    """
    client: "AbstractAndesiteWebSocketClient"
    guild_id: int
    op: str
    body: Dict[str, Any]

    def __init__(self, client: "AbstractAndesiteWebSocketClient", guild_id: int, op: str, body: Dict[str, Any]) -> None:
        super().__init__(client=client, guild_id=guild_id, op=op, body=body)



"""Combined client for Andesite.

This client combines the `AndesiteWebSocket` and `AndesiteHTTP` clients.

There is the `AndesiteClientBase` which is simply the implementation
of `AbstractAndesiteWebSocket` and `AbstractAndesiteHTTP` and then there
is the actual client `AndesiteClient`.

You can use the combined client for everything that implements
the abstract methods. This means that you can use the combined client
for `ClientPool` clients!
"""

import asyncio
from contextlib import suppress
from typing import Any, Dict, Optional, Union

from yarl import URL

from .event_target import EventTarget
from .http_client import AbstractAndesiteHTTP, AndesiteHTTPBase, AndesiteHTTPInterface
from .state import AbstractAndesiteState, _get_state
from .web_socket_client import AbstractAndesiteWebSocket, AbstractAndesiteWebSocketClient, AndesiteWebSocketBase, \
    AndesiteWebSocketInterface

__all__ = ["AndesiteClientBase", "AndesiteClient", "create_andesite_client"]


class AndesiteClientBase(AbstractAndesiteWebSocket, AbstractAndesiteHTTP, EventTarget):
    """Implementation of `AbstractAndesiteWebSocket` and `AbstractAndesiteHTTP`.

    See Also:
        Use the `create` class method if you want to conveniently create the client.

    Args:
        http_client: Client to use for the http endpoints.
        web_socket_client: Client to use for the web socket communication.
        loop: Event loop to use. This is used for the underlying event target.

    The `event_target` of the web socket client is set to the combined client.

    This class implements `AbstractAndesiteWebSocketClient` if the underlying web socket client
    is an instance of it. However, an `isinstance` check will always return `False`.
    If the underlying client is not an instance of `AbstractAndesiteWebSocketClient`, all
    related methods will return `None`.

    Attributes:
        http (AbstractAndesiteHTTP): HTTP client which is used for the `AbstractAndesiteHTTP` methods.
        web_socket (AbstractAndesiteWebSocket): WebSocket client which is used for the `AbstractAndesiteWebSocket` methods.
    """

    http: AbstractAndesiteHTTP
    web_socket: AbstractAndesiteWebSocket

    def __init__(self, http_client: AbstractAndesiteHTTP, web_socket_client: AbstractAndesiteWebSocket, *,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        super().__init__(loop=loop)

        self.http = http_client
        self.web_socket = web_socket_client

        # set the web socket client's event target to our event target
        # self.event_target will most likely resolve to self, but maybe
        # the user is trying to pull some inception shit with nested
        # clients and who would I be to deny them their fun?
        web_socket_client.event_target.add_dispatcher(self.event_target)

        # bind the methods directly
        self.send = web_socket_client.send
        self.request = http_client.request

    def __repr__(self) -> str:
        return f"{type(self).__name__}(http_client={self.http!r}, web_socket_client={self.web_socket!r})"

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.http}, {self.web_socket})"

    @property
    def closed(self) -> bool:
        """Whether this client is closed and can no longer be used.

        This is `True` if either of the underlying clients is closed.
        """
        return self.http.closed or self.web_socket.closed

    @property
    def state(self) -> Optional[AbstractAndesiteState]:
        return self.web_socket.state

    @state.setter
    def state(self, value: Optional[AbstractAndesiteState]) -> None:
        self.web_socket.state = value

    @property
    def connected(self) -> Optional[bool]:
        """Whether the web socket client is connected and usable.

        This method returns `None` if the web socket client isn't a
        `AbstractAndesiteWebSocketClient`.
        """
        if not isinstance(self.web_socket, AbstractAndesiteWebSocketClient):
            return None

        return self.web_socket.connected

    @property
    def connection_id(self) -> Optional[str]:
        """Connection id of the web socket client.

        This method returns `None` if the web socket client isn't a
        `AbstractAndesiteWebSocketClient`.
        """
        if not isinstance(self.web_socket, AbstractAndesiteWebSocketClient):
            return None

        return self.web_socket.connection_id

    async def connect(self, *, max_attempts: int = None) -> None:
        """Connect the underlying web socket client.

        Args:
            max_attempts: Amount of connection attempts to perform before aborting.
                If `None`, unlimited attempts will be performed.

        This method doesn't do anything if the web socket client isn't a
        `AbstractAndesiteWebSocketClient`.
        """
        if not isinstance(self.web_socket, AbstractAndesiteWebSocketClient):
            return

        await self.web_socket.connect(max_attempts=max_attempts)

    async def disconnect(self) -> None:
        """Disconnect the underlying web socket client.

        This method doesn't do anything if the web socket client isn't a
        `AbstractAndesiteWebSocketClient`.
        """
        if not isinstance(self.web_socket, AbstractAndesiteWebSocketClient):
            return

        await self.web_socket.disconnect()

    async def send(self, guild_id: int, op: str, payload: Dict[str, Any]) -> None:
        """Send a message using the web socket client.

        See Also:
            Refer to `AbstractAndesiteWebSocket.send` for the documentation.
        """
        await self.web_socket.send(guild_id, op, payload)

    async def close(self) -> None:
        """Close the underlying clients.

        This closes both the http and the web socket client.
        Since the `close` method shouldn't raise an exception anyway,
        all exceptions are suppressed. This ensures that both
        clients' close methods are called.
        """
        with suppress(Exception):
            await self.http.close()

        with suppress(Exception):
            await self.web_socket.close()

    async def reset(self) -> None:
        """Reset the underlying clients so they may be used again.

        This has the opposite effect of the `close` method making the clients
        usable again.
        """
        await asyncio.gather(self.http.reset(), self.web_socket.reset())

    async def request(self, method: str, path: str, **kwargs) -> Any:
        """Perform a request on the http client.

        See Also:
            Refer to `AbstractAndesiteHTTP.request` for the documentation.
        """
        return await self.http.request(method, path, **kwargs)


class AndesiteClient(AndesiteWebSocketInterface, AndesiteHTTPInterface, AndesiteClientBase):
    """Andesite client which combines the web socket with the http client.

    See Also:
        Use the `create` class method if you want to conveniently create the client.

    Args:
        http_client: Client to use for the http endpoints.
        web_socket_client: Client to use for the web socket communication.

    This class implements `AbstractAndesiteWebSocketClient` if the underlying web socket client
    is an instance of it. However, an `isinstance` check will always return `False`.
    If the underlying client is not an instance of `AbstractAndesiteWebSocketClient`, all
    related methods will return `None`.

    Attributes:
        http (AbstractAndesiteHTTP): HTTP client which is used for the `AbstractAndesiteHTTP` methods.
        web_socket (AbstractAndesiteWebSocket): WebSocket client which is used for the `AbstractAndesiteWebSocket` methods.
    """
    ...


def create_andesite_client(http_uri: Union[str, URL], web_socket_uri: Union[str, URL], password: Optional[str],
                           user_id: int, *,
                           state: Union[AbstractAndesiteState, bool] = None,
                           loop: asyncio.AbstractEventLoop = None):
    """Create a new combined Andesite client.

    Args:
        http_uri: URI for the http endpoint
        web_socket_uri: URI for the web socket endpoint
        password: Andesite password for authorization
        user_id: User ID
        state: State handler to use. Defaults to in-memory `AndesiteState`.
            You can pass `False` to disable state handling.
        loop: Specify the event loop to use. The
            meaning of this value depends on the client,
            but you can safely omit it.

    Returns:
        A new combined client with `AndesiteHTTPBase` and `AndesiteWebSocketBase` as
        its clients.
    """

    http_client = AndesiteHTTPBase(http_uri, password, loop=loop)
    web_socket_client = AndesiteWebSocketBase(web_socket_uri, user_id, password, loop=loop)

    inst = AndesiteClient(http_client, web_socket_client, loop=loop)

    inst.state = _get_state(state)
    return inst



"""Operations sent by Andesite.

Attributes:
    EVENT_MAP (Mapping[str, Type[AndesiteEvent]]): Mapping from the event name
        to the corresponding `AndesiteEvent` type. See: `get_event_model`
    OP_MAP (Mapping[str, Type[ReceiveOperation]]): Mapping from the op code to
    the corresponding `ReceiveOperation` type. See: `get_update_model`

See Also:
    `andesite.models.send_operations` for operations sent to Andesite.
"""

import abc
import copy
import dataclasses
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Mapping, Optional, Set, TYPE_CHECKING, Tuple, Type, Union, cast

from andesite.event_target import NamedEvent
from andesite.transform import RawDataType, map_build_values_from_raw, map_convert_values, \
    map_convert_values_from_milli, map_convert_values_to_milli, map_remove_keys, transform_input, transform_output
from .debug import Error, Stats
from .player import Player

if TYPE_CHECKING:
    from andesite import AbstractAndesiteWebSocketClient

__all__ = ["ReceiveOperation",
           "ConnectionUpdate", "StatsUpdate", "PlayerUpdate",
           "AndesiteEvent", "TrackStartEvent",
           "TrackEndReason", "TrackEndEvent",
           "TrackExceptionEvent", "TrackStuckEvent", "UnknownAndesiteEvent",
           "get_event_model", "get_update_model"]


class ReceiveOperation(abc.ABC):
    """Message sent by Andesite.

    Attributes:
        client (Optional[AbstractAndesiteWebSocketClient]): Client that received
            the message. This is set by the client that received the message.
    """
    __op__: str

    client: Optional["AbstractAndesiteWebSocketClient"] = None


@dataclass
class PongResponse(ReceiveOperation):
    """Simple pong response sent as a response to ping requests.

    Attributes:
        user_id (int): User id
        guild_id (int): Guild id
    """
    __op__ = "pong"

    user_id: int
    guild_id: int

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class ConnectionUpdate(ReceiveOperation):
    """Message sent upon connecting to the Web socket.

    Attributes:
        id (str): Connection ID
    """
    __op__ = "connection-id"

    id: str


@dataclass
class MetadataUpdate(ReceiveOperation):
    """Payload sent on connection start containing handshake response header.

    Attributes:
        data (Dict[str, Union[int, str, List[str]]]): Map of metadata key to
            value.
    """
    __op__ = "metadata"

    data: Dict[str, Union[int, str, List[str]]]


@dataclass
class StatsUpdate(ReceiveOperation):
    """Message containing statistics.

    Attributes:
        user_id (int): User ID
        stats (Stats): Statistics
    """
    __op__ = "stats"

    user_id: int
    stats: Stats

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int)
        map_build_values_from_raw(data, stats=Stats)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str)


@dataclass
class PlayerUpdate(NamedEvent, ReceiveOperation):
    """Player update sent by Andesite for active players.

    Attributes:
        user_id (int): user id
        guild_id (int): guild id
        state (Player): player
    """
    __op__ = "player-update"

    user_id: int
    guild_id: int
    state: Player

    def __post_init__(self) -> None:
        super().__init__()

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)
        map_build_values_from_raw(data, state=Player)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class AndesiteEvent(NamedEvent, ReceiveOperation, abc.ABC):
    """Event sent by Andesite.

    Attributes:
        type (str): Event type name.
            This is equal to the name of the class (With the exception of
            `UnknownAndesiteEvent`)
        user_id (int): User ID
        guild_id (int): Guild ID
        track (str): Base64 encoded track data

    Inherits from `NamedEvent` so it can be dispatched by an `EventTarget`.
    """
    __op__ = "event"
    __event_name__ = "andesite_event"

    type: str
    user_id: int
    guild_id: int

    def __post_init__(self) -> None:
        super().__init__()

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class TrackStartEvent(AndesiteEvent):
    """Event emitted when a new track starts playing."""

    track: str


class TrackEndReason(Enum):
    """Reason why a track stopped playing.

    See Also:
        `TrackEndEvent`

    Attributes:
        FINISHED: Usually caused by the track reaching the end,
            however it will also be used when it ends due to an exception.
        LOAD_FAILED: Track failed to start, throwing an exception before
            providing any audio.
        STOPPED: Track was stopped due to the player being stopped.
        REPLACED: Track stopped playing because a new track started playing.
        CLEANUP: Track was stopped because the cleanup threshold for the audio
            player was reached.
    """
    FINISHED = "FINISHED"
    LOAD_FAILED = "LOAD_FAILED"
    STOPPED = "STOPPED"
    REPLACED = "REPLACED"
    CLEANUP = "CLEANUP"


@dataclass
class TrackEndEvent(AndesiteEvent):
    """Event emitted when a track ended.

    Attributes:
        reason (TrackEndReason): Reason why a track stopped playing.
        may_start_next (bool): Indicates whether a new track should be started
            on receiving this event. If this is `False`, either this event is
            already triggered because another track started
            (`TrackEndReason.REPLACED`) or because the player is stopped
            (`TrackEndReason.STOPPED`, `TrackEndReason.CLEANUP`).
    """

    track: str
    reason: TrackEndReason
    may_start_next: bool

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        data = transform_input(super(), data)
        map_convert_values(data, reason=TrackEndReason)
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)
        data["reason"] = cast(TrackEndReason, data["reason"]).value
        return data


@dataclass
class TrackExceptionEvent(AndesiteEvent):
    """Event emitted when there's an error.

    Attributes:
        error (str): Error message
        exception (Error): Error data
    """
    track: str

    error: str
    exception: Error


@dataclass
class TrackStuckEvent(AndesiteEvent):
    """Event emitted when a track is stuck.

    Attributes:
        threshold (float): Threshold in seconds
    """

    track: str
    threshold: float

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        data = transform_input(super(), data)
        map_convert_values_from_milli(data, "threshold")
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)
        map_convert_values_to_milli(data, "threshold")
        return data


@dataclass
class WebSocketClosedEvent(AndesiteEvent):
    """Event emitted when the Andesite node disconnects from a voice channel.

    Attributes:
        reason (str): Reason for the disconnect
        code (int): Error code
        by_remote (bool): Whether the disconnect was caused by the remote.
    """

    reason: str
    code: int
    by_remote: bool


@dataclass
class UnknownAndesiteEvent(AndesiteEvent):
    """Special kind of event for unknown events.

    This shouldn't occur at all unless the library is out-dated.

    Attributes:
        body (object): Entire event body sent by Andesite.
            Please note that the keys are in snake_case.

    """
    __event_name__ = "unknown_andesite_event"

    body: RawDataType

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        # we want a clean body... does that sound weird? no, I'm sure it doesn't
        data["body"] = copy.deepcopy(data)

        map_remove_keys(data, *(key for key in data.keys() if key not in _UNKNOWN_ANDESITE_EVENT_FIELD_NAMES))
        data = transform_input(super(), data)
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)

        body: RawDataType = data.pop("body")
        # let the existing data overwrite the body
        body.update(data)
        return body


_UNKNOWN_ANDESITE_EVENT_FIELDS: Tuple[dataclasses.Field] = dataclasses.fields(UnknownAndesiteEvent)
_UNKNOWN_ANDESITE_EVENT_FIELD_NAMES: Set[str] = {field.name for field in _UNKNOWN_ANDESITE_EVENT_FIELDS}

_EVENTS: Set[Type[AndesiteEvent]] = {
    TrackStartEvent,
    TrackEndEvent,
    TrackExceptionEvent, TrackStuckEvent,
    WebSocketClosedEvent,
}
EVENT_MAP: Mapping[str, Type[AndesiteEvent]] = {event.__name__: event for event in _EVENTS}


def get_event_model(event_type: str) -> Type[AndesiteEvent]:
    """Get the model corresponding to the given event name.

    Args:
        event_type: Event type name.

    Returns:
        Model type for the given event type.
        `UnknownAndesiteEvent` if no matching event was found.
    """
    try:
        return EVENT_MAP[event_type]
    except KeyError:
        return UnknownAndesiteEvent


_OPS: Set[Type[ReceiveOperation]] = {PongResponse, ConnectionUpdate, MetadataUpdate, StatsUpdate, PlayerUpdate}
OP_MAP: Mapping[str, Type[ReceiveOperation]] = {op.__op__: op for op in _OPS}


def get_update_model(op: str, event_type: str = None) -> Optional[Type[ReceiveOperation]]:
    """Get the model corresponding to the given op code.

    Args:
        op: Op code sent by Andesite
        event_type: Event type if and only if op is "event".
            This is used to return the correct event type.
            See `get_event_model`.
            If not set and op is "event" the function returns `None`.
    """
    if op == "event":
        if event_type is None:
            return None

        return get_event_model(event_type)
    else:
        return OP_MAP.get(op)



"""Andesite audio filters.

Attributes:
    FILTER_MAP (Mapping[str, Type[Filter]]): Mapping from filter name to filter class.
        See: `get_filter_model`.
    FilterMapLike (Union[FilterMap, Dict[str, Union[Filter, RawDataType]]]): (Type alias) Type of objects which
        can be used as filter maps. This includes the `FilterMap`.
"""
import abc
from dataclasses import dataclass, field
from operator import eq
from typing import Any, Dict, Iterable, Iterator, List, Mapping, MutableMapping, Optional, Set, Type, TypeVar, Union, \
    overload

from andesite.transform import RawDataType, build_from_raw, convert_to_raw

__all__ = ["Filter",
           "EqualizerBand", "Equalizer",
           "Karaoke",
           "Timescale",
           "Tremolo",
           "Vibrato",
           "VolumeFilter",
           "get_filter_model",
           "FilterMap", "FilterMapLike"]


def _ensure_in_interval(value: float, *,
                        low: float = None, low_inc: float = None,
                        up: float = None, up_inc: float = None) -> None:
    """Ensure a value is within an interval.

    Raises:
        ValueError: If the provided value isn't within the given
            constraints.
    """
    low_symbol: Optional[str] = None
    up_symbol: Optional[str] = None
    valid: bool = True

    if low_inc is not None:
        low_symbol = f"[{low_inc}"
        if not value >= low:
            valid = False
    elif low is not None:
        low_symbol = f"({low}"
        if not value > low:
            valid = False

    if up_inc is not None:
        up_symbol = f"{up_inc}]"
        if not value <= up_inc:
            valid = False
    elif up is not None:
        up_symbol = f"{up})"
        if value < up_inc:
            valid = False

    if not valid:
        low_symbol = low_symbol or "[-INF"
        up_symbol = up_symbol or "INF]"
        raise ValueError(f"Provided value ({value}) not in interval {low_symbol}, {up_symbol}!")


class _Filter(abc.ABC):
    """Filter with name.

    Attributes:
        __filter_name__ (str): Name of the filter.
            This is a magic attribute used by the library
            to convert the filter into its Andesite representation.
    """
    __filter_name__: str


@dataclass
class Filter(_Filter, abc.ABC):
    """Audio filter for Andesite.

    Attributes:
        enabled (bool): Whether or not the filter is enabled.
            This value is mostly useful when receiving the filters from Andesite.
            However you can also set it to `False` when sending filters. This
            will cause the settings to be ignored and instead the default values
            are sent to Andesite which will cause the filter to be disabled.

    When creating a new `Filter` instance its values are set to the
    default value.
    """
    enabled: bool = True

    def reset(self) -> None:
        """Reset the filter settings back to their default values."""
        self.__init__()

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        enabled = data.pop("enabled")
        if enabled:
            return data
        else:
            # create a new instance (which uses the defaults) and return its data
            return convert_to_raw(cls())


@dataclass
class EqualizerBand:
    """

    Attributes:
        band (int): band number to configure ( 0 - 14 )
        gain (float): value to set for the band ( [-0.25, 1.0] )
    """
    band: int
    gain: float = 0.0

    def set_band(self, value: int) -> None:
        """Setter for :py:attr:`band` which performs a value check.

        Args:
            value: Value to set for the band. ( [0, 14] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low_inc=0, up_inc=14)
        self.band = value

    def set_gain(self, value: float) -> None:
        """Setter for :py:attr:`gain` which performs a value check.

        Args:
            value: Value to set for the gain. ( [-0.25, 1] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low_inc=-.25, up_inc=1)
        self.gain = value


@dataclass
class Equalizer(Filter):
    """

    Attributes:
        bands (List[EqualizerBand]): array of bands to configure
    """
    __filter_name__ = "equalizer"

    bands: List[EqualizerBand] = field(default_factory=list)

    def __iter__(self) -> Iterator[EqualizerBand]:
        return iter(self.bands)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Equalizer):
            for my_gain, other_gain in zip(self.iter_band_gains(), other.iter_band_gains()):
                if my_gain != other_gain:
                    return False

            return True
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.iter_band_gains())

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        # Andesite sends the equalizer filter as an array of floats
        bands = data["bands"]
        for i, gain in enumerate(bands):
            bands[i] = EqualizerBand(i, gain)

    @classmethod
    def from_gains(cls, gains: Iterable[Optional[float]]) -> "Equalizer":
        """Create an `Equalizer` filter from a list of gains.

        Args:
            gains: Iterable of `float` which correspond to the gain for the band,
                or `None` if the band doesn't specify a gain.
        """
        bands: List[EqualizerBand] = []
        for i, gain in enumerate(gains):
            if gain is not None:
                bands.append(EqualizerBand(i, gain))

        return cls(True, bands)

    @overload
    def get_band(self, band: int) -> EqualizerBand:
        ...

    @overload
    def get_band(self, band: int, create: bool) -> Optional[EqualizerBand]:
        ...

    def get_band(self, band: int, create: bool = True) -> Optional[EqualizerBand]:
        """Get the specified band from the bands list.

        If the band doesn't exist it is created.
        If you don't want to automatically create a band, pass
        `create=False`.

        Args:
            band: Band number to get
            create: Whether or not to create a new band if it doesn't exist. (Defaults to True)
        """
        try:
            return next(band for band in self.bands if band.band == band)
        except StopIteration:
            pass

        if not create:
            return None

        band = EqualizerBand(band)
        self.bands.append(band)
        return band

    def get_band_gain(self, band: int) -> Optional[float]:
        """Get the gain of a band.

        Returns:
            Gain of the band or `None` if it doesn't exist.
        """
        band = self.get_band(band, create=False)
        if band:
            return band.gain
        else:
            return None

    def set_band_gain(self, band: int, gain: float) -> None:
        """Set the gain of a band to the specified value.

        If the band does not exist it is created.

        Args:
            band: Band number to set the gain for.
            gain: Value to set for the gain. ( [-0.25, 1] )

        Raises:
            ValueError: if the provided gain is invalid.
        """
        self.get_band(band).set_gain(gain)

    @overload
    def iter_band_gains(self, use_default: bool) -> List[Optional[float]]:
        ...

    @overload
    def iter_band_gains(self) -> List[float]:
        ...

    def iter_band_gains(self, use_default: bool = True) -> List[Optional[float]]:
        """Get a list of all the bands' gains in order.

        Args:
            use_default: Whether or not to replace non-existent values
                with the default gain.
                If `False` and band doesn't have a gain set, `None`
                is used instead.
        """
        default_value: Union[float, None] = EqualizerBand.gain if use_default else None

        gains: List[Optional[float]] = 15 * [default_value]
        for band in self:
            gain = band.gain
            if use_default and gain is None:
                continue

            gains[band.band] = gain

        return gains


@dataclass
class Karaoke(Filter):
    """

    Attributes:
        level (float)
        mono_level (float)
        filter_band (float)
        filter_width (float)
    """
    __filter_name__ = "karaoke"

    level: float = 1.0
    mono_level: float = 1.0
    filter_band: float = 220.0
    filter_width: float = 220.0


@dataclass
class Timescale(Filter):
    """

    Attributes:
        speed (float): speed to play music at (> 0)
        pitch (float): pitch to set (> 0)
        rate (float): rate to set (> 0)
    """
    __filter_name__ = "timescale"

    speed: float = 1.0
    pitch: float = 1.0
    rate: float = 1.0

    def set_speed(self, value: float) -> None:
        """Setter for :py:attr:`speed` which performs a value check.

        Args:
            value: Value to set for the speed. ( (0, INF] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0)
        self.speed = value

    def set_pitch(self, value: float) -> None:
        """Setter for :py:attr:`pitch` which performs a value check.

        Args:
            value: Value to set for the pitch. ( (0, INF] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0)
        self.pitch = value

    def set_rate(self, value: float) -> None:
        """Setter for :py:attr:`rate` which performs a value check.

        Args:
            value: Value to set for the rate. ( (0, INF] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0)
        self.rate = value


@dataclass
class Tremolo(Filter):
    """

    Attributes:
        frequency (float): (> 0)
        depth (float): ( (0, 1] )
    """
    __filter_name__ = "tremolo"

    frequency: float = 2.0
    depth: float = 0.5

    def set_frequency(self, value: float) -> None:
        """Setter for :py:attr:`frequency` which performs a value check.

        Args:
            value: Value to set for the frequency. ( (0, INF] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0)
        self.frequency = value

    def set_depth(self, value: float) -> None:
        """Setter for :py:attr:`depth` which performs a value check.

        Args:
            value: Value to set for the depth. ( (0, 1] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0, up_inc=1)
        self.depth = value


@dataclass
class Vibrato(Filter):
    """

    Attributes:
        frequency (float): ( (0, 14] )
        depth (float): ( (0, 1] )
    """
    __filter_name__ = "vibrato"

    frequency: float = 2.0
    depth: float = 0.5

    def set_frequency(self, value: float) -> None:
        """Setter for :py:attr:`frequency` which performs a value check.

        Args:
            value: Value to set for the frequency. ( (0, 14] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0, up_inc=14)
        self.frequency = value

    def set_depth(self, value: float) -> None:
        """Setter for :py:attr:`depth` which performs a value check.

        Args:
            value: Value to set for the depth. ( (0, 1] )

        Raises:
            ValueError: if the provided value is invalid.
        """
        _ensure_in_interval(value, low=0, up_inc=1)
        self.depth = value


@dataclass
class VolumeFilter(Filter):
    """Volume filter settings.

    Attributes:
        volume (float): Volume modifier. This acts as a factor for the actual volume.
    """
    __filter_name__ = "volume"

    volume: float = 1.0


_FILTERS: Set[Type[Filter]] = {Equalizer, Karaoke, Timescale, Tremolo, Vibrato, VolumeFilter}
FILTER_MAP: Mapping[str, Type[Filter]] = {andesite_filter.__filter_name__: andesite_filter for andesite_filter in
                                          _FILTERS}


def get_filter_model(name: str) -> Optional[Type[Filter]]:
    """Get the corresponding filter model for the given name.

    If no model for the name exists, `None` is returned.

    Args:
        name: Name of the filter
    """
    return FILTER_MAP.get(name)


FT = TypeVar("FT", bound=Filter)


@dataclass
class FilterMap(MutableMapping):
    """Custom mapping type for filters.

    Attributes:
        filters (Dict[str, Any]): Dictionary containing all filters.

    Theoretically this is just a wrapper around the `filters` dictionary which
    contains the actual filter data. The class exposes the known filters as
    properties, but it also supports unknown filters should the library
    become outdated.

    You can also use this as a wrapper for an existing filter dict.
    """

    filters: Dict[str, Any] = field(default_factory=dict)

    def __init__(self, filters: "FilterMapLike") -> None:
        if isinstance(filters, FilterMap):
            self.filters = filters.filters.copy()
        else:
            self.filters = filters

    def __eq__(self, other) -> bool:
        if isinstance(other, FilterMap):
            return eq(self.filters, other.filters)
        elif isinstance(other, Mapping):
            return eq(self.filters, other)
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.filters)

    def __len__(self) -> int:
        return len(self.filters)

    def __iter__(self) -> Iterator[str]:
        return iter(self.filters)

    def __getitem__(self, item: str) -> Any:
        return self.filters[item]

    def __setitem__(self, key: str, value: Any) -> None:
        self.filters[key] = value

    def __delitem__(self, key: str) -> None:
        del self.filters[key]

    def get_filter(self, name: str, cls: Type[FT]) -> FT:
        """Get the filter with the name.

        Args:
            name: Name of the filter to get
            cls: `Filter` class to use for the filter.
        """
        try:
            value = self[name]
        except KeyError:
            value = self[name] = cls()
        else:
            if not isinstance(value, cls):
                if isinstance(value, dict):
                    value = self[name] = cls(**value)
                else:
                    raise TypeError(f"Expected {cls}, found {type(value)!r}: {value}")

        return value

    @property
    def equalizer(self) -> Equalizer:
        """Equalizer filter settings"""
        return self.get_filter("equalizer", Equalizer)

    @property
    def karaoke(self) -> Karaoke:
        """Karaoke filter settings"""
        return self.get_filter("karaoke", Karaoke)

    @property
    def timescale(self) -> Timescale:
        """Timescale filter settings"""
        return self.get_filter("timescale", Timescale)

    @property
    def tremolo(self) -> Tremolo:
        """Tremolo filter settings"""
        return self.get_filter("tremolo", Tremolo)

    @property
    def vibrato(self) -> Vibrato:
        """Vibrato filter settings"""
        return self.get_filter("vibrato", Vibrato)

    @property
    def volume(self) -> VolumeFilter:
        """Volume filter settings"""
        return self.get_filter("equalizer", VolumeFilter)

    def set_filter(self, andesite_filter: Filter) -> None:
        """Set the value for a filter.

        Args:
            andesite_filter: Filter to set
        """
        self[andesite_filter.__filter_name__] = andesite_filter

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        filters: RawDataType = {}

        for name, filter_value in data.items():
            filter_cls = get_filter_model(name)

            if filter_cls is not None:
                filter_value = build_from_raw(filter_cls, filter_value)

            filters[name] = filter_value

        return dict(filters=filters)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        return data["filters"]


FilterMapLike = Union[FilterMap, Dict[str, Union[Filter, RawDataType]]]



"""Track models.

Attributes:
    UNKNOWN_ARTIST (str): Value used by Lavaplayer if an artist is unknown.
"""

from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from andesite.transform import RawDataType, build_from_raw, map_build_values_from_raw, map_convert_value, \
    map_convert_values_from_milli, map_convert_values_to_milli, seq_build_all_items_from_raw
from .debug import Error

__all__ = ["PlaylistInfo", "TrackMetadata", "TrackInfo", "LoadType", "LoadedTrack"]


@dataclass
class PlaylistInfo:
    """Information for of the playlist a track belongs to.

    Attributes:
        name: name of the playlist
        selected_track: index of the selected track in the tracks list,
            or `None` if no track is selected
    """
    name: str
    selected_track: Optional[int]

    def __str__(self) -> str:
        res: List[str] = [self.name]

        if self.selected_track is not None:
            res.append(f"[{self.selected_track}]")

        return " ".join(res)


UNKNOWN_ARTIST: str = "Unknown artist"


@dataclass
class TrackMetadata:
    """Additional information for a track.

    Attributes:
        class_name (str): class name of the lavaplayer track
        title (str): title of the track
        author (str): author of the track
        length (Optional[float]): duration of the track, in seconds.
            Set to `None` if it's a stream. Andesite would normally return
            2^63 - 1 (Long.MAX_VALUE), but that value doesn't have any
            significance in Python so it's represented with something more
            meaningful.
        identifier (str): identifier of the track
        uri (str): uri of the track
        is_stream (bool): whether or not the track is a livestream
        is_seekable (bool): whether or not the track supports seeking
        position (float): current position of the track
    """
    class_name: str
    title: str
    author: str
    length: Optional[float]
    identifier: str
    uri: str
    is_stream: bool
    is_seekable: bool
    position: float

    def __str__(self) -> str:
        return f"{self.author} - {self.title}"

    @property
    def author_unknown(self) -> bool:
        """Whether or not the author is unknown.

        Uses constant `UNKNOWN_ARTIST` to check against.

        Keep in mind that there is no way to detect whether
        the author's name just happens to coincide with the
        `UNKNOWN_ARTIST` text. Use with caution.
        """
        return self.author == UNKNOWN_ARTIST

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        data["class_name"] = data.pop("class")

        # set length to None if we're dealing with a stream.
        try:
            is_stream = data["is_stream"]
        except KeyError:
            pass
        else:
            if is_stream:
                data["length"] = None

        map_convert_values_from_milli(data, "length", "position")

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        data["class"] = data.pop("class_name")
        map_convert_values_to_milli(data, "length", "position")


@dataclass
class TrackInfo:
    """Track info represents a track and its metadata.

    Attributes:
        track (str): base64 encoded track
        info (TrackMetadata): metadata of the track
    """
    track: str
    info: TrackMetadata

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_build_values_from_raw(data, info=TrackMetadata)


class LoadType(Enum):
    """Load type of a `LoadedTrack`"""
    TRACK_LOADED = "TRACK_LOADED"
    SEARCH_RESULT = "SEARCH_RESULT"
    PLAYLIST_LOADED = "PLAYLIST_LOADED"
    NO_MATCHES = "NO_MATCHES"
    LOAD_FAILED = "LOAD_FAILED"

    @property
    def success(self) -> bool:
        """Whether the load type is a successful one.

        Only `LOAD_FAILED` counts as unsuccessful.
        """
        return self != LoadType.LOAD_FAILED


@dataclass
class LoadedTrack:
    """Result provided by a load track request.

    Attributes:
        load_type (LoadType): type of the response
        tracks (Optional[List[TrackInfo]]): loaded tracks
        playlist_info (Optional[PlaylistInfo]): metadata of the loaded playlist
        cause (Optional[Error]): error that happened while loading tracks
        severity (Optional[str]): severity of the error

    This class provides the magic methods `__bool__` and `__len__` which
    operate in respect to `tracks`.
    """
    load_type: LoadType
    tracks: Optional[List[TrackInfo]]
    playlist_info: Optional[PlaylistInfo]

    cause: Optional[Error] = None
    severity: Optional[str] = None

    def __bool__(self) -> bool:
        """Return `True` if tracks have been loaded, `False` otherwise.

        This doesn't check the `load_type`, instead it simply returns the
        `bool` value of `tracks`.
        """
        return bool(self.tracks)

    def __len__(self) -> int:
        """Return the amount of tracks.

        Returns:
            Amount of tracks loaded, if `tracks` is `None` (i.e. no tracks have been loaded)
            it returns 0.
        """
        if self.tracks is None:
            return 0
        else:
            return len(self.tracks)

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_value(data, "load_type", LoadType)

        try:
            playlist_info = data["playlist_info"]
        except KeyError:
            pass
        else:
            # my god, Andesite now sends an empty object if the playlist_info is null
            # and we absolutely don't want that! So if it's falsy, use None
            if playlist_info:
                playlist_info = build_from_raw(PlaylistInfo, playlist_info)
            else:
                playlist_info = None

            data["playlist_info"] = playlist_info

        map_build_values_from_raw(data, cause=Error)

        try:
            tracks = data["tracks"]
        except KeyError:
            pass
        else:
            seq_build_all_items_from_raw(tracks, TrackInfo)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        data["load_type"] = data["load_type"].value

    def get_selected_track(self) -> Optional[TrackInfo]:
        """Get the selected track.

        "Selected" means that either the result is a playlist with an
        entry selected (`PlaylistInfo.selected_track`), or the result
        is a `LoadType.TRACK_LOADED` result (in which case the first and only
        track is returned).

        All other cases will return in `None` being returned.

        Returns:
            The selected track if there is any, or `None`
            if the selected track doesn't exist.
        """
        if self.tracks is None:
            return None

        if self.playlist_info:
            selected_track = self.playlist_info.selected_track
        elif self.load_type == LoadType.TRACK_LOADED:
            selected_track = 0
        else:
            selected_track = None

        if selected_track is None:
            return None

        try:
            return self.tracks[selected_track]
        except IndexError:
            return None



"""State handler for Andesite clients.

If you want to store your state externally for example in a database, there
are two approaches you can use. You can either implement the
`AbstractAndesiteState` and use the default `AndesitePlayerState` (which can be
easily converted to and from JSON), or you can use the default `AndesiteState`
with a custom `AbstractPlayerState` implementation.
Both approaches have their advantages, but you should always consider that while
the get operations are often performed together (especially during state
migration), the set operations are not.

Attributes:
    StateArgumentType (Union[AbstractAndesiteState, bool, None]): (Type alias)
        types which can be passed as a state handler to a client.
"""

import abc
import asyncio
import functools
import logging
from typing import Any, Awaitable, Callable, Coroutine, Dict, Generic, Optional, TypeVar, Union

from .event_target import NamedEvent
from .models import AndesiteEvent, Player, PlayerUpdate, TrackEndEvent, TrackExceptionEvent, TrackStartEvent, \
    TrackStuckEvent, VoiceServerUpdate
from .transform import build_from_raw, convert_to_raw

__all__ = ["AbstractPlayerState", "PlayerState",
           "AbstractAndesiteState", "AndesiteState",
           "player_to_raw", "player_from_raw",
           "voice_server_update_to_raw", "voice_server_update_from_raw",
           "StateArgumentType", "_get_state"]

log = logging.getLogger(__name__)


def player_to_raw(player: Player) -> Dict[str, Any]:
    """Convert the given player to a JSON-serialisable dict.

    Args:
        player: Player to convert.

    Returns:
        A JSON-Serialisable dict containing all the data of the player.
        This data can be passed to `player_from_raw` which creates a `Player`
        instance again.
    """
    return convert_to_raw(player)


def player_from_raw(data: Dict[str, Any]) -> Player:
    """Recreate a player from the converted dict.

    Args:
        data: Data returned by `player_to_raw` to convert to a Player.

    Raises:
          Exception: If the data can't be used to create a Player.

    Returns:
        Created `Player` instance.
    """
    return build_from_raw(Player, data)


def voice_server_update_to_raw(update: VoiceServerUpdate) -> Dict[str, Any]:
    """Convert the given voice server update to a JSON-serialisable dict.

    Args:
        update: Voice server update to convert.

    Returns:
        A JSON-Serialisable dict containing all the data of the voice server
        update. This data can be passed to `voice_server_update_from_raw` which
        creates a `VoiceServerUpdate` instance again.
    """
    return convert_to_raw(update)


def voice_server_update_from_raw(data: Dict[str, Any]) -> VoiceServerUpdate:
    """Recreate a voice server update from the converted dict.

    Args:
        data: Data returned by `voice_server_update_to_raw` to convert.

    Raises:
          Exception: If the data can't be used to create a voice server update.

    Returns:
        Created `VoiceServerUpdate` instance.
    """
    return build_from_raw(VoiceServerUpdate, data)


class AbstractPlayerState(abc.ABC):
    """State of a single Andesite player.

    Notes:
        Unless you're doing something weird you don't need to call the setter
        functions. If the player state is managed by an `AbstractAndesiteState`
        which is connected to a client then everything is done for you.

    See Also:
        `AndesitePlayerState` for an in-memory implementation.
    """

    def __str__(self) -> str:
        return f"PlayerState(guild_id={self.guild_id})"

    @property
    @abc.abstractmethod
    def guild_id(self) -> int:
        """ID of the guild the player is for."""
        ...

    @abc.abstractmethod
    async def get_player(self) -> Optional[Player]:
        """Get the player information.

        Returns:
            `Player` information or `None` if no player information
            is present.
        """
        ...

    @abc.abstractmethod
    async def set_player(self, player: Optional[Player]) -> None:
        """Set the current player.

        Args:
            player: Player data to set. May be `None` to remove the player
                information.
        """
        ...

    @abc.abstractmethod
    async def get_voice_server_update(self) -> Optional[VoiceServerUpdate]:
        """Get the last voice server update that was sent to the player.

        Returns:
            The last voice server update or `None` if none exists.
        """
        ...

    @abc.abstractmethod
    async def set_voice_server_update(self, update: Optional[VoiceServerUpdate]) -> None:
        """Set the last voice server update.

        Args:
            update: Voice server update to set.
        """
        ...

    @abc.abstractmethod
    async def get_track(self) -> Optional[str]:
        """Get the currently playing track."""
        ...

    @abc.abstractmethod
    async def set_track(self, track: Optional[str]) -> None:
        """Set the currently playing track."""
        ...


class PlayerState(AbstractPlayerState):
    """Default player state storing the state in memory.

    The player state can be converted to a JSON-serialisable dict using
    `to_raw`. A state can also be created from said dict using the classmethod
    `from_raw`. These methods exist to make it easy to implement a custom
    `AbstractAndesiteState` which loads and stores serialised player states.
    """

    def __init__(self, guild_id: int):
        self._guild_id = guild_id

        self._player = None
        self._track = None
        self._voice_server_update = None

    @property
    def guild_id(self) -> int:
        return self._guild_id

    async def get_player(self) -> Optional[Player]:
        return self._player

    async def set_player(self, player: Optional[Player]) -> None:
        self._player = player

    async def get_track(self) -> Optional[str]:
        return self._track

    async def set_track(self, track: Optional[str]) -> None:
        self._track = track

    async def get_voice_server_update(self) -> Optional[VoiceServerUpdate]:
        return self._voice_server_update

    async def set_voice_server_update(self, update: Optional[VoiceServerUpdate]) -> None:
        self._voice_server_update = update

    @classmethod
    def from_raw(cls, data: Dict[str, Any]) -> AbstractPlayerState:
        """Create an `AbstractPlayerState` from the raw data.

        Args:
            data: Raw player state data as returned by `to_raw`

        Returns:
            A new instance of `PlayerState` describing the same state
            as the data.
        """
        inst = cls(data["guild_id"])

        inst._track = data.get("track")

        try:
            raw_player = data["player"]
        except KeyError:
            pass
        else:
            if raw_player is not None:
                inst._player = player_from_raw(raw_player)

        try:
            raw_voice_server_update = data["raw_voice_server_update"]
        except KeyError:
            pass
        else:
            if raw_voice_server_update is not None:
                inst._voice_server_update = voice_server_update_from_raw(raw_voice_server_update)

        return inst

    def to_raw(self) -> Dict[str, Any]:
        """Convert the player state into a JSON-serialisable dict.

        Use the classmethod `from_raw` to re-create a `PlayerState` using the
        returned data.
        """
        if self._player:
            raw_player = player_to_raw(self._player)
        else:
            raw_player = None

        if self._voice_server_update:
            raw_voice_server_update = voice_server_update_to_raw(self._voice_server_update)
        else:
            raw_voice_server_update = None

        return {
            "guild_id": self._guild_id,
            "player": raw_player,
            "track": self._track,
            "voice_server_update": raw_voice_server_update,
        }


def _run_with_error_callback(coro: Coroutine, err_cb: Callable[[Exception], Awaitable], *,
                             loop: asyncio.AbstractEventLoop = None) -> asyncio.Future:
    async def wrapper() -> None:
        try:
            await coro
        except Exception as e:
            await err_cb(e)

    return asyncio.ensure_future(wrapper(), loop=loop)


class AbstractAndesiteState(abc.ABC):
    """Andesite state handler.

    Keeps track of the state of an Andesite node.
    """

    def __str__(self) -> str:
        return f"{type(self).__name__}"

    def _handle_andesite_message(self, message: NamedEvent, *,
                                 loop: asyncio.AbstractEventLoop = None) -> Optional[asyncio.Future]:
        """Handles the event of an andesite message being received.

        Args:
            message: Message that was sent.
            loop: Event loop to use.
        """
        if isinstance(message, PlayerUpdate):
            coro = self.handle_player_update(message)
        elif isinstance(message, AndesiteEvent):
            coro = self.handle_andesite_event(message)
        else:
            return None

        err_cb: Callable = functools.partial(self.on_handle_message_error, message)
        return _run_with_error_callback(coro, err_cb, loop=loop)

    def _handle_sent_message(self, guild_id: int, op: str, payload: Dict[str, Any], *,
                             loop: asyncio.AbstractEventLoop = None) -> Optional[asyncio.Future]:
        """Handles the event of a message being sent.

        Args:
            guild_id: Guild id the message was sent for.
            op: Operation code.
            payload: Raw payload of the message.
            loop: Event loop to use.
        """
        if op == "voice-server-update":
            update = VoiceServerUpdate(payload["sessionId"], payload["event"])
            coro = self.handle_voice_server_update(guild_id, update)
        else:
            return None

        err_cb: Callable = functools.partial(self.on_handle_sent_message_error, guild_id, op, payload)
        return _run_with_error_callback(coro, err_cb, loop=loop)

    @abc.abstractmethod
    async def handle_player_update(self, update: PlayerUpdate) -> None:
        """Handle a player update.

        Args:
            update: Update that was received.
        """
        ...

    @abc.abstractmethod
    async def handle_andesite_event(self, event: AndesiteEvent) -> None:
        """Handle an Andesite event.

        Args:
            event: Andesite event that was received.
        """
        ...

    @abc.abstractmethod
    async def handle_voice_server_update(self, guild_id: int, update: VoiceServerUpdate) -> None:
        """Handle a voice server update.

        Args:
            guild_id: Guild the update applies to.
            update: Voice server update.
        """
        ...

    async def on_handle_message_error(self, message: NamedEvent, exc: Exception) -> None:
        """Called when an error occurs during message handling.

        Args:
            message: Message that caused the error
            exc: Exception that was raised.
        """
        log.error(f"uncaught error {exc} in {self} when handling message {message}")

    async def on_handle_sent_message_error(self, guild_id: int, op: str, payload: Dict[str, Any],
                                           exc: Exception) -> None:
        """Called when an error occurs during sent message handling.

        Args:
            guild_id: Guild the message was sent for.
            op: Operation code.
            payload: Raw payload which was sent.
            exc: Exception that was raised.
        """
        log.error(f"uncaught error {exc} in {self} when handling sent message {op} for guild {guild_id}: {exc}\n\n"
                  f"Payload: {payload}")

    @abc.abstractmethod
    async def get(self, guild_id: int) -> AbstractPlayerState:
        """Get the player state of a guild.

        Args:
            guild_id: Guild to get player state for.

        Returns:
            Player state for the given guild.
        """
        ...


PST = TypeVar("PST", bound=AbstractPlayerState)


class AndesiteState(AbstractAndesiteState, Generic[PST]):
    """Default implementation of `AbstractAndesiteState`.

    Stores the player states in memory.

    Args:
        state_factory: Callable which when called with a guild id, creates an
            `AbstractPlayerState`. The default is `PlayerState`.

    Attributes:
        player_states (Dict[int, AbstractPlayerState]): Mapping from guild id
            to the corresponding player state.
    """
    player_states: Dict[int, PST]
    _state_factory: Callable[[int], PST]

    def __init__(self, *, state_factory: Callable[[int], PST] = PlayerState) -> None:
        self.player_states = {}

        self._state_factory = state_factory

    def __repr__(self) -> str:
        return f"{type(self).__name__}(state_factory={self._state_factory!r})"

    def _get_or_create_player_state(self, guild_id: int) -> PST:
        try:
            player_state = self.player_states[guild_id]
        except KeyError:
            player_state = self.player_states[guild_id] = self._state_factory(guild_id)

        return player_state

    async def handle_player_update(self, update: PlayerUpdate) -> None:
        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"handling player update for {self}: {update}")

        state = self._get_or_create_player_state(update.guild_id)
        await state.set_player(update.state)

    async def handle_andesite_event(self, event: AndesiteEvent) -> None:
        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"handling andesite event for {self}: {event}")

        if isinstance(event, (TrackEndEvent, TrackExceptionEvent, TrackStuckEvent)):
            track = None
        elif isinstance(event, TrackStartEvent):
            track = event.track
        else:
            return

        state = self._get_or_create_player_state(event.guild_id)
        await state.set_track(track)

    async def handle_voice_server_update(self, guild_id: int, update: VoiceServerUpdate) -> None:
        player_state = self._get_or_create_player_state(guild_id)
        await player_state.set_voice_server_update(update)

    async def get(self, guild_id: int) -> PST:
        return self._get_or_create_player_state(guild_id)


StateArgumentType = Union[AbstractAndesiteState, bool, None]


def _get_state(state: StateArgumentType) -> Optional[AbstractAndesiteState]:
    """Handle state creation/suppression.

    Args:
        state: State which was passed to the function.

    Raises:
        TypeError: If an invalid state argument was passed.

    Returns:
        An instance of `AndesiteState` if state is `None` or `True`,
        `None` if state is `False`, and `state` itself if it's a state.
    """
    if state is None or state is True:
        return AndesiteState()
    elif state is False:
        return None
    elif isinstance(state, AbstractAndesiteState):
        return state
    else:
        raise TypeError("State must implement AbstractAndesiteState. "
                        "You can also use False to disable state handling or"
                        "None to use the default state handler.")



"""A client library for Andesite.

The top level `andesite` namespace also
exports all the models from `andesite.models`
and the various clients.

Attributes:
    __version__: Version of the installed andesite.py library.
"""

__version__ = "0.1.2"

from .combined_client import *
from .http_client import *
from .pools import *
from .state import *
from .web_socket_client import *
from .web_socket_client_events import *

# must come after other imports!
from .models import *



"""Debug models for Andesite.

These models are used in either `Stats` which represents the Andesite
stats returned by `AndesiteWebSocketInterface.get_stats` or `Error` which is
used to represent an Andesite error.
"""

from dataclasses import dataclass
from typing import List, NoReturn, Optional

from andesite.transform import RawDataType, map_build_values_from_raw, map_convert_values, map_rename_keys, \
    seq_build_all_items_from_raw

__all__ = ["StackFrame", "Error", "AndesiteException",
           "PlayersStats", "RuntimeVMStats", "RuntimeSpecStats", "RuntimeVersionStats", "RuntimeStats", "OSStats", "CPUStats", "ClassLoadingStats",
           "ThreadStats", "CompilationStats", "MemoryCommonUsageStats", "MemoryStats", "GCStats", "MemoryPoolStats", "MemoryManagerStats",
           "FrameStats",
           "Stats"]


@dataclass
class StackFrame:
    """Andesite stack frame.

    Can be found in `Error.stack`.

    Attributes:
        class_loader (Optional[str]): name of the classloader
        module_name (Optional[str]): name of the module
        module_version (Optional[str]): version of the module
        class_name (str): name of the class
        method_name (str): name of the method
        file_name (Optional[str]): name of the source file
        line_number (Optional[int]): line in the source file
        pretty (str): pretty printed version of this frame, as it would appear on Throwable#printStackTrace
    """
    class_loader: Optional[str]
    module_name: Optional[str]
    module_version: Optional[str]
    class_name: str
    method_name: str
    file_name: Optional[str]
    line_number: Optional[int]
    pretty: str

    def __str__(self) -> str:
        spec: List[str] = []

        for name, value in [("module", self.module_name), ("file", self.file_name), ("line", self.line_number)]:
            # only show 'em if we know 'em
            if value is not None:
                spec.append(f"{name} {value}")

        spec.append(f"{self.class_name}#{self.method_name}")

        return ", ".join(spec)


@dataclass
class Error:
    """Andesite error.

    You can convert the Andesite error data into a Python exception
    using the `as_python_exception` method and the `raise_python_exception` to
    raise it.

    Attributes:
        class_name (str): class of the error
        message (Optional[str]): message of the error
        stack (List[StackFrame]): stacktrace of the error
        cause (Optional[Error]): cause of the error
        suppressed (List[Error]): suppressed errors
    """
    class_name: str
    message: Optional[str]
    stack: List[StackFrame]
    suppressed: List["Error"]
    cause: Optional["Error"]

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        seq_build_all_items_from_raw(data["stack"], StackFrame)
        seq_build_all_items_from_raw(data["suppressed"], Error)
        map_build_values_from_raw(data, cause=Error)
        map_rename_keys(data, class_name="class")

    def as_python_exception(self) -> "AndesiteException":
        """Create an `AndesiteException` which can be raised.

        If `cause` is not `None` it is added to the exception.
        """
        exc = AndesiteException(self.class_name, self.message, self.stack, self.suppressed)

        if self.cause is not None:
            exc.__cause__ = self.cause.as_python_exception()

        return exc

    def raise_error(self) -> NoReturn:
        """Raise the Andesite error as an `AndesiteException`.

        Raises:
            AndesiteException: Generated using the `as_python_exception` method.
        """
        raise self.as_python_exception()


class AndesiteException(Exception):
    """Andesite's `Error` represented as python exceptions.

    Attributes:
        class_name (str): Class name of the error
        message (Optional[str]): message of the error
        stack (List[StackFrame]): cause of the error
        suppressed (List[Error]): suppressed errors
    """

    class_name: str
    message: Optional[str]
    stack: List[StackFrame]
    suppressed: List[Error]

    def __init__(self, class_name: str, message: Optional[str], stack: List[StackFrame], suppressed: List[Error]) -> None:
        super().__init__(class_name, message)
        self.class_name = class_name
        self.message = message
        self.stack = stack
        self.suppressed = suppressed

    def __str__(self) -> str:
        return f"{self.class_name}: {self.message}"


# STATISTICS


@dataclass
class PlayersStats:
    """Players statistics sent by Andesite.

    Attributes:
        total (int): Total amount of players
        playing (int): Amount of players that are actively playing

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    total: int
    playing: int


@dataclass
class RuntimeVMStats:
    """VM statistics.
    
    Attributes:
        name (str)
        vendor (str)
        version (str)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    vendor: str
    version: str


@dataclass
class RuntimeSpecStats:
    """Spec statistics.
    
    Attributes:
        name (str)
        vendor (str)
        version (str)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    vendor: str
    version: str


@dataclass
class RuntimeVersionStats:
    """Version information stats.
    
    Attributes:
        feature (int)
        interim (int)
        update (int)
        patch (int)
        pre (Optional[str])
        build (int)
        optional (str)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    feature: int
    interim: int
    update: int
    patch: int
    pre: Optional[str]
    build: int
    optional: str


@dataclass
class RuntimeStats:
    """Runtime statistics.
    
    Attributes:
        uptime (int)
        pid (int)
        management_spec_version (str)
        name (str)
        vm (RuntimeVMStats)
        spec (RuntimeSpecStats)
        version (RuntimeVersionStats)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    uptime: int
    pid: int
    management_spec_version: str
    name: str
    vm: RuntimeVMStats
    spec: RuntimeSpecStats
    version: RuntimeVersionStats

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_build_values_from_raw(data, vm=RuntimeVMStats, spec=RuntimeSpecStats, version=RuntimeVersionStats)


@dataclass
class OSStats:
    """OS statistics.
    
    Attributes:
        processors (int)
        name (str)
        arch (str)
        version (str)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    processors: int
    name: str
    arch: str
    version: str


@dataclass
class CPUStats:
    """CPU statistics.

    Attributes:
        andesite (float)
        system (float)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    andesite: float
    system: float


@dataclass
class ClassLoadingStats:
    """Class loading statistics.
    
    Attributes:
        loaded (int)
        total_loaded (int)
        unloaded (int)
    """
    loaded: int
    total_loaded: int
    unloaded: int


@dataclass
class ThreadStats:
    """Thread statistics.
    
    Attributes:
        running (int)
        daemon (int)
        peak (int)
        total_started (int)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    running: int
    daemon: int
    peak: int
    total_started: int


@dataclass
class CompilationStats:
    """Compilation statistics.
    
    Attributes:
        name (str)
        total_time (int)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    total_time: int


@dataclass
class MemoryCommonUsageStats:
    """Memory usage statistics.
    
    Attributes:
        init (int)
        used (int)
        committed (int)
        max (int)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    init: int
    used: int
    committed: int
    max: int


@dataclass
class MemoryStats:
    """Memory statistics.

    Attributes:
        pending_finalization (int)
        heap (MemoryCommonUsageStats)
        non_heap (MemoryCommonUsageStats)

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    pending_finalization: int
    heap: MemoryCommonUsageStats
    non_heap: MemoryCommonUsageStats

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_build_values_from_raw(data, heap=MemoryCommonUsageStats, non_heap=MemoryCommonUsageStats)


@dataclass
class GCStats:
    """Garbage collection statistics.

    Attributes:
        name (str)
        collection_count (int)
        collection_time (int)
        pools (List[str])

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    collection_count: int
    collection_time: int
    pools: List[str]


@dataclass
class MemoryPoolStats:
    """Memory pool statistics.

    Attributes:
        name (str): Name of the pool
        type (str): Type of the pool. This is either "HEAP" or "NON_HEAP".
        collection_usage (Optional[MemoryHeapStats])
        collection_usage_threshold (Optional[int])
        collection_usage_threshold_count (Optional[int])
        peak_usage (MemoryCommonUsageStats)
        usage (MemoryCommonUsageStats)
        usage_threshold (int)
        usage_threshold_count (int)
        managers (List[str])

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    type: str
    collection_usage: Optional[MemoryCommonUsageStats]
    collection_usage_threshold: Optional[int]
    collection_usage_threshold_count: Optional[int]
    peak_usage: MemoryCommonUsageStats
    usage: MemoryCommonUsageStats
    usage_threshold: int
    usage_threshold_count: int
    managers: List[str]

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_build_values_from_raw(data, collection_usage=MemoryCommonUsageStats, peak_usage=MemoryCommonUsageStats, usage=MemoryCommonUsageStats)


@dataclass
class MemoryManagerStats:
    """Memory manager statistics.

    Attributes:
        name (str): Name of the manager
        pools (List[str]): Memory pools

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    name: str
    pools: List[str]


@dataclass
class FrameStats:
    """Frame statistics for a guild player.

    Attributes:
        user (int): User ID
        guild (int): Guild ID
        success (int): Amount of successful frames
        loss (int): Amount of lost frames

    See Also:
        This statistic can be found in `Stats` which is retrieved from Andesite by
        the clients. Both `AndesiteHTTP` and `AndesiteWebSocket` are able to get them.
    """
    user: int
    guild: int
    success: int
    loss: int

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user=int, guild=int)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user=str, guild=str)


@dataclass
class Stats:
    """Statistics sent by Andesite

    Attributes:
        players (PlayersStats): Player statistics
        runtime (RuntimeStats): Runtime statistics
        os (OSStats): OS statistics
        cpu (CPUStats): CPU statistics
        class_loading (ClassLoadingStats): Class loading statistics
        thread (ThreadStats): Thread statistics
        compilation (CompilationStats): Compilation statistics
        memory (MemoryStats): Memory statistics
        gc (List[GCStats]): GC statistics
        memory_pools (List[MemoryPoolStats]): Memory pool statistics
        memory_managers (List[MemoryManagerStats]): Memory manager statistics
        frame_stats (List[FrameStats]): Frame statistics
    """

    players: PlayersStats
    runtime: RuntimeStats
    os: OSStats
    cpu: CPUStats
    class_loading: ClassLoadingStats
    thread: ThreadStats
    compilation: CompilationStats
    memory: MemoryStats
    gc: List[GCStats]
    memory_pools: List[MemoryPoolStats]
    memory_managers: List[MemoryManagerStats]
    frame_stats: List[FrameStats]

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_build_values_from_raw(data,
                                  players=PlayersStats, runtime=RuntimeStats, os=OSStats, cpu=CPUStats, class_loading=ClassLoadingStats,
                                  thread=ThreadStats, compilation=CompilationStats, memory=MemoryStats,
                                  )
        seq_build_all_items_from_raw(data["gc"], GCStats)
        seq_build_all_items_from_raw(data["memory_pools"], MemoryPoolStats)
        seq_build_all_items_from_raw(data["memory_managers"], MemoryManagerStats)
        seq_build_all_items_from_raw(data["frame_stats"], FrameStats)



from pathlib import Path

from setuptools import find_packages, setup

import andesite

long_description = Path("README.md").read_text()

setup(
    name="andesite.py",
    version=andesite.__version__,
    description="Andesite client for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Giesela Inc.",
    packages=find_packages(exclude=("examples", "docs", "tests")),
    python_requires=">=3.7",
    install_requires=[
        "aiohttp",
        "lettercase",
        "websockets",
        "yarl",
    ],
)
